# -*- coding: utf-8 -*-

import fcntl
import os
import shutil
import subprocess

from .config import Config
from .paths import HOSTS_FILE, HOSTS_LOCK, LXC_ROOT
from .templates import LXC_CONTAINER

def prepare_container(layers):
    # Extract the variables from values given via lxc.hook.pre-start hook
    app = os.environ['LXC_NAME']
    # Remove ephemeral layer data
    clean_ephemeral_layer(app)
    # Prepare and mount overlayfs
    prepare_overlayfs(app, layers)
    # Configure host and common params used in the app
    configure_app(app)

def clean_ephemeral_layer(app):
    # Cleans containers ephemeral layer. Called in lxc.hook.post-stop and lxc.hook.pre-start in case of unclean shutdown
    # This is done early in the container start process, so the inode of the ephemeral directory must remain unchanged
    ephemeral = os.path.join(LXC_ROOT, app, 'ephemeral')
    for item in os.scandir(ephemeral):
        shutil.rmtree(item.path) if item.is_dir() else os.unlink(item.path)

def prepare_overlayfs(app, layers):
    # Prepare and mount overlayfs
    rootfs = os.path.join(LXC_ROOT, app, 'rootfs')
    # Unmount rootfs in case it remained mounted for whatever reason
    subprocess.run(['umount', rootfs])
    layers = layers.split(',')
    if len(layers) == 1:
        # We have only single layer, no overlay needed
        subprocess.run(['mount', '--bind', layers[0], rootfs])
    else:
        olwork = os.path.join(LXC_ROOT, app, 'olwork')
        subprocess.run(['mount', '-t', 'overlay', '-o', 'upperdir={},lowerdir={},workdir={}'.format(layers[-1], ','.join(layers[:-1]), olwork), 'none', rootfs])

def configure_app(app):
    # Supply common configuration for the application. Done as part of container preparation during service startup
    script = os.path.join('/srv', app, 'update-conf.sh')
    if os.path.exists(script):
        conf = Config()
        setup_env = os.environ.copy()
        setup_env['DOMAIN'] = conf['host']['domain']
        setup_env['PORT'] = conf['host']['port']
        setup_env['EMAIL'] = conf['common']['email']
        setup_env['GMAPS_API_KEY'] = conf['common']['gmaps-api-key']
        subprocess.run([script], env=setup_env, check=True)

def cleanup_container():
    # Extract the variables from values given via lxc.hook.post-stop hook
    app = os.environ['LXC_NAME']
    # Unmount rootfs
    rootfs = os.path.join(LXC_ROOT, app, 'rootfs')
    subprocess.run(['umount', rootfs])
    # Remove ephemeral layer data
    clean_ephemeral_layer(app)

def register_container(app, image):
    # Create directories after container installation
    rootfs = os.path.join(LXC_ROOT, app, 'rootfs')
    olwork = os.path.join(LXC_ROOT, app, 'olwork')
    ephemeral = os.path.join(LXC_ROOT, app, 'ephemeral')
    os.makedirs(rootfs, 0o755, True)
    os.makedirs(olwork, 0o755, True)
    os.makedirs(ephemeral, 0o755, True)
    os.chown(ephemeral, 100000, 100000)
    # Create container configuration file
    layers = ','.join([os.path.join(LXC_ROOT, 'storage', layer) for layer in image['layers']])
    if 'build' not in image:
        layers = '{},{}'.format(layer, ephemeral)
    mounts = '\n'.join(['lxc.mount.entry = {} {} none bind,create={} 0 0'.format(m[1], m[2], m[0].lower()) for m in image['mounts']]) if 'mounts' in image else ''
    env = '\n'.join(['lxc.environment = {}={}'.format(e[0], e[1]) for e in image['env']]) if 'env' in image else ''
    uid = image['uid'] if 'uid' in image else '0'
    gid = image['gid'] if 'gid' in image else '0'
    cmd = image['cmd'] if 'cmd' in image else '/bin/sh'
    cwd = image['cwd'] if 'cwd' in image else '/'
    halt = image['halt'] if 'halt' in image else 'SIGINT'
    # Lease the first unused IP to the container
    ipv4 = update_hosts_lease(app, True)
    # Create the config file
    with open(os.path.join(LXC_ROOT, app, 'config'), 'w') as f:
        f.write(LXC_CONTAINER.format(name=app, ipv4=ipv4, layers=layers, mounts=mounts, env=env, uid=uid, gid=gid, cmd=cmd, cwd=cwd, halt=halt))

def unregister_container(app):
    # Remove container configuration and directories
    # TODO: Duplicated with what pkgmgr does, ale zustane to tady, protoze unregister se pouziva pri buildu
    shutil.rmtree(os.path.join(LXC_ROOT, app))
    # Release the IP address
    update_hosts_lease(app, False)

def update_hosts_lease(app, is_request):
    # This is a poor man's DHCP server which uses /etc/hosts as lease database
    # Leases the first unused IP from range 172.17.0.0/16
    # Uses file lock as interprocess mutex
    ip = None
    with open(HOSTS_LOCK, 'w') as lock:
        fcntl.lockf(lock, fcntl.LOCK_EX)
        # Load all existing records
        with open(HOSTS_FILE, 'r') as f:
            leases = [l.strip().split(' ', 1) for l in f]
        # If this call is a request for lease, find the first unassigned IP
        if is_request:
            used_ips = [l[0] for l in leases]
            for i in range(2, 65278): # Reserve last /24 subnet for VPN
                ip = '172.17.{}.{}'. format(i // 256, i % 256)
                if ip not in used_ips:
                    leases.append([ip, app])
                    break
        # Otherwise it is a release in which case we just delete the record
        else:
            leases = [l for l in leases if l[1] != app]
        # Write the contents back to the file
        with open(HOSTS_FILE, 'w') as f:
            for lease in leases:
                f.write('{} {}\n'.format(lease[0], lease[1]))
    return ip
